# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import importlib
import logging
import os

from oslo_config import cfg
from oslo_log import log

import tobiko

LOG = log.getLogger(__name__)

CONFIG_MODULES = ['tobiko.openstack.keystone.config',
                  'tobiko.openstack.neutron.config',
                  'tobiko.openstack.nova.config',
                  'tobiko.shell.ssh.config',
                  'tobiko.shell.ping.config',
                  'tobiko.shell.sh.config']

CONFIG_DIRS = [os.getcwd(),
               os.path.expanduser("~/.tobiko"),
               '/etc/tobiko']


class GlobalConfig(object):

    # this is a singletone
    _instance = None
    _sources = {}

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def set_source(self, name, source_conf):
        if source_conf is None:
            raise TypeError("Config source cannot be None")
        actual = self._sources.setdefault(name, source_conf)
        if actual is not source_conf:
            msg = "Config source already registered: {!r}".format(name)
            raise RuntimeError(msg)

    def __getattr__(self, name):
        sources = self._sources.get(name)
        if sources is None:
            msg = "Config source not registered: {!r}".format(name)
            raise NoSuchConfigSource(msg)
        return sources


CONF = GlobalConfig()


def init_config():
    init_tobiko_config()
    init_environ_config()


def init_tobiko_config(default_config_dirs=None, product_name='tobiko',
                       version='unknown'):
    default_config_dirs = default_config_dirs or CONFIG_DIRS

    # Register configuration options
    conf = cfg.ConfigOpts()
    log.register_options(conf)
    register_tobiko_options(conf=conf)

    # Initialize tobiko configuration object
    conf(args=[], default_config_dirs=default_config_dirs)
    CONF.set_source('tobiko', conf)

    # setup final configuration
    log.setup(conf=conf, product_name=product_name, version=version)
    setup_tobiko_config(conf=conf)


def register_tobiko_options(conf):

    conf.register_opts(
        group=cfg.OptGroup('http'),
        opts=[cfg.StrOpt('http_proxy',
                         help="HTTP proxy URL for Rest APIs"),
              cfg.StrOpt('https_proxy',
                         help="HTTPS proxy URL for Rest APIs"),
              cfg.StrOpt('no_proxy',
                         help="Don't use proxy server to connect to listed "
                         "hosts")])

    for module_name in CONFIG_MODULES:
        module = importlib.import_module(module_name)
        if hasattr(module, 'register_tobiko_options'):
            module.register_tobiko_options(conf=conf)


def setup_tobiko_config(conf):
    # Redirect all warnings to logging library
    logging.captureWarnings(True)
    warnings_logger = log.getLogger('py.warnings')
    if conf.debug:
        if not warnings_logger.isEnabledFor(log.WARNING):
            # Print Python warnings
            warnings_logger.logger.setLevel(log.WARNING)
    elif warnings_logger.isEnabledFor(log.WARNING):
        # Silence Python warnings
        warnings_logger.logger.setLevel(log.ERROR)

    tobiko.setup_fixture(HttpProxyFixture)

    for module_name in CONFIG_MODULES:
        module = importlib.import_module(module_name)
        if hasattr(module, 'setup_tobiko_config'):
            module.setup_tobiko_config(conf=conf)


class HttpProxyFixture(tobiko.SharedFixture):
    """Make sure we have http proxy environment variables defined when required
    """

    http_proxy = None
    https_proxy = None
    no_proxy = None
    source = None

    def setup_fixture(self):
        source = None
        http_proxy = os.environ.get('http_proxy')
        https_proxy = os.environ.get('https_proxy')
        no_proxy = os.environ.get('no_proxy')
        if http_proxy or https_proxy:
            source = 'environment'
        else:
            http_conf = CONF.tobiko.http
            http_proxy = http_conf.http_proxy
            https_proxy = http_conf.https_proxy
            no_proxy = http_conf.no_proxy
            if http_proxy:
                os.environ['http_proxy'] = http_proxy
            if https_proxy:
                os.environ['https_proxy'] = https_proxy
            if http_proxy or https_proxy:
                source = 'tobiko.conf'
                if no_proxy:
                    os.environ['no_proxy'] = no_proxy

        if source:
            LOG.info("Using HTTP proxy configuration defined in %s:\n"
                     "  http_proxy: %r\n"
                     "  https_proxy: %r\n"
                     "  no_proxy: %r",
                     source, os.environ.get('http_proxy'),
                     os.environ.get('https_proxy'), os.environ.get('no_proxy'))
        else:
            LOG.debug("Connecting to REST API services without a proxy "
                      "server")

        self.source = source
        self.http_proxy = os.environ.get('http_proxy')
        self.https_proxy = os.environ.get('https_proxy')
        self.no_proxy = os.environ.get('no_proxy')


def init_environ_config():
    CONF.set_source('environ', EnvironConfig())


class EnvironConfig(object):

    def __getattr__(self, name):
        value = get_env(name)
        if value is None:
            msg = "Environment variable not defined: {!r}".format(name)
            raise cfg.NoSuchOptError(msg)
        return value


class NoSuchConfigSource(AttributeError):
    pass


def get_any_option(*sources, **kwargs):
    default = kwargs.get('default', None)
    for source in sources:
        value = CONF
        for name in source.split('.'):
            try:
                value = getattr(value, name)
            except (NoSuchConfigSource,
                    cfg.NoSuchOptError,
                    cfg.NoSuchGroupError) as ex:
                LOG.debug("No such option value for %r: %s", source, ex)
                break
        else:
            if value != default:
                return value
    return default


def get_env(name):
    value = os.environ.get(name)
    if value:
        return value
    else:
        LOG.debug("Environment variable %r is not defined")
        return None


def get_int_env(name):
    value = get_env(name)
    if value:
        try:
            return int(value)
        except TypeError:
            LOG.exception("Environment variable %r is not an integer: %r",
                          name, value)
    return None


init_config()



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import select
import subprocess
import sys
import time

from oslo_log import log
import six

import tobiko
from tobiko.shell import ssh
from tobiko.shell.sh import _exception


LOG = log.getLogger(__name__)


def execute(command, stdin=None, environment=None, timeout=None, shell=None,
            check=True, ssh_client=None):
    """Execute command inside a remote or local shell

    :param command: command argument list

    :param timeout: command execution timeout in seconds

    :param check: when False it doesn't raises ShellCommandError when
    exit status is not zero. True by default

    :param ssh_client: SSH client instance used for remote shell execution

    :returns: STDOUT text when command execution terminates with zero exit
    status.

    :raises ShellTimeoutExpired: when timeout expires before command execution
    terminates. In such case it kills the process, then it eventually would
    try to read STDOUT and STDERR buffers (not fully implemented) before
    raising the exception.

    :raises ShellCommandError: when command execution terminates with non-zero
    exit status.
    """

    if timeout:
        timeout = float(timeout)

    ssh_client = ssh_client or ssh.ssh_proxy_client()
    if not ssh_client and not shell:
        from tobiko import config
        CONF = config.CONF
        shell = CONF.tobiko.shell.command

    if shell:
        command = split_command(shell) + [join_command(command)]
    else:
        command = split_command(command)

    if ssh_client:
        result = execute_remote_command(command=command, stdin=stdin,
                                        environment=environment,
                                        timeout=timeout,
                                        ssh_client=ssh_client)
    else:
        result = execute_local_command(command=command, stdin=stdin,
                                       environment=environment,
                                       timeout=timeout)

    if result.exit_status == 0:
        LOG.debug("Command %r succeeded:\n"
                  "stderr:\n%s\n"
                  "stdout:\n%s\n",
                  command, result.stderr, result.stdout)
    elif result.exit_status is None:
        LOG.debug("Command %r timeout expired (timeout=%s):\n"
                  "stderr:\n%s\n"
                  "stdout:\n%s\n",
                  command, timeout, result.stderr, result.stdout)
    else:
        LOG.debug("Command %r failed (exit_status=%s):\n"
                  "stderr:\n%s\n"
                  "stdout:\n%s\n",
                  command, result.exit_status, result.stderr, result.stdout)
    if check:
        result.check()

    return result


def execute_remote_command(command, ssh_client, stdin=None, timeout=None,
                           environment=None):
    """Execute command on a remote host using SSH client"""

    if isinstance(ssh_client, ssh.SSHClientFixture):
        # Connect to fixture
        ssh_client = tobiko.setup_fixture(ssh_client).client

    transport = ssh_client.get_transport()
    with transport.open_session() as channel:
        if environment:
            channel.update_environment(environment)
        channel.exec_command(join_command(command))
        stdout, stderr = comunicate_ssh_channel(channel, stdin=stdin,
                                                timeout=timeout)
        if channel.exit_status_ready():
            exit_status = channel.recv_exit_status()
        else:
            exit_status = None
    return ShellExecuteResult(command=command, timeout=timeout,
                              stdout=stdout, stderr=stderr,
                              exit_status=exit_status)


def comunicate_ssh_channel(ssh_channel, stdin=None, chunk_size=None,
                           timeout=None, sleep_time=None, read_stdout=True,
                           read_stderr=True):
    if read_stdout:
        rlist = [ssh_channel]
    else:
        rlist = []

    if not stdin:
        ssh_channel.shutdown_write()
        stdin = None
        wlist = []
    else:
        wlist = [ssh_channel]
        if not isinstance(stdin, six.binary_type):
            stdin = stdin.encode()

    chunk_size = chunk_size or 1024
    sleep_time = sleep_time or 1.
    timeout = timeout or float("inf")
    start = time.time()
    stdout = None
    stderr = None

    while True:
        chunk_timeout = min(sleep_time, timeout - (time.time() - start))
        if chunk_timeout < 0.:
            LOG.debug('Timed out reading from SSH channel: %r', ssh_channel)
            break
        ssh_channel.settimeout(chunk_timeout)
        if read_stdout and ssh_channel.recv_ready():
            chunk = ssh_channel.recv(chunk_size)
            if stdout:
                stdout += chunk
            else:
                stdout = chunk
            if not chunk:
                LOG.debug("STDOUT channel closed by peer on SSH channel %r",
                          ssh_channel)
                read_stdout = False
        elif read_stderr and ssh_channel.recv_stderr_ready():
            chunk = ssh_channel.recv_stderr(chunk_size)
            if stderr:
                stderr += chunk
            else:
                stderr = chunk
            if not chunk:
                LOG.debug("STDERR channel closed by peer on SSH channel %r",
                          ssh_channel)
                read_stderr = False
        elif ssh_channel.exit_status_ready():
            break
        elif stdin and ssh_channel.send_ready():
            sent_bytes = ssh_channel.send(stdin[:chunk_size])
            stdin = stdin[sent_bytes:] or None
            if not stdin:
                LOG.debug('shutdown_write() on SSH channel: %r', ssh_channel)
                ssh_channel.shutdown_write()
        else:
            select.select(rlist, wlist, rlist or wlist, chunk_timeout)

    if stdout:
        if not isinstance(stdout, six.string_types):
            stdout = stdout.decode()
    else:
        stdout = ''
    if stderr:
        if not isinstance(stderr, six.string_types):
            stderr = stderr.decode()
    else:
        stderr = ''
    return stdout, stderr


def execute_local_command(command, stdin=None, environment=None, timeout=None):
    """Execute command on local host using local shell"""

    LOG.debug("Executing command %r on local host (timeout=%r)...",
              command, timeout)

    stdin = stdin or None
    process = subprocess.Popen(command,
                               universal_newlines=True,
                               env=environment,
                               stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)

    if timeout and sys.version_info < (3, 3):
        LOG.warning("Popen.communicate method doens't support for timeout "
                    "on Python %r", sys.version)
        timeout = None

    # Wait for process execution while reading STDERR and STDOUT streams
    if timeout:
        try:
            # pylint: disable=unexpected-keyword-arg,no-member
            stdout, stderr = process.communicate(input=stdin,
                                                 timeout=timeout)
        except subprocess.TimeoutExpired:
            # At this state I expect the process to be still running
            # therefore it has to be kill later after calling poll()
            LOG.exception("Command %r timeout expired.", command)
            stdout = stderr = ''
    else:
        stdout, stderr = process.communicate(input=stdin)

    # Check process termination status
    exit_status = process.poll()
    if exit_status is None:
        # The process is still running after calling communicate():
        # let kill it
        process.kill()

    return ShellExecuteResult(command=command, timeout=timeout,
                              stdout=stdout, stderr=stderr,
                              exit_status=exit_status)


class ShellExecuteResult(collections.namedtuple(
        'ShellExecuteResult', ['command', 'timeout', 'exit_status', 'stdout',
                               'stderr'])):

    def check(self):
        command = join_command(self.command)
        if self.exit_status is None:
            raise _exception.ShellTimeoutExpired(command=command,
                                                 timeout=self.timeout,
                                                 stderr=self.stderr,
                                                 stdout=self.stdout)

        elif self.exit_status != 0:
            raise _exception.ShellCommandFailed(command=command,
                                                exit_status=self.exit_status,
                                                stderr=self.stderr,
                                                stdout=self.stdout)


def split_command(command):
    if isinstance(command, six.string_types):
        return command.split()
    elif command:
        return [str(a) for a in command]
    else:
        return []


def join_command(command):
    if isinstance(command, six.string_types):
        return command
    elif command:
        return subprocess.list2cmdline([str(a) for a in command])
    else:
        return ""



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import os

from oslo_log import log
import paramiko

import tobiko


LOG = log.getLogger(__name__)


def ssh_config(config_files=None):
    if config_files:
        fixture = SSHConfigFixture(config_files=config_files)
    else:
        fixture = SSHConfigFixture
    return tobiko.setup_fixture(fixture)


def ssh_host_config(host=None, config_files=None):
    return ssh_config(config_files=config_files).lookup(host)


class SSHDefaultConfigFixture(tobiko.SharedFixture):

    conf = None

    def setup_fixture(self):
        from tobiko import config
        CONF = config.CONF
        self.conf = CONF.tobiko.ssh

    def __getattr__(self, name):
        return getattr(self.conf, name)


class SSHConfigFixture(tobiko.SharedFixture):

    default = tobiko.required_setup_fixture(SSHDefaultConfigFixture)

    config_files = None
    config = None

    def __init__(self, config_files=None):
        super(SSHConfigFixture, self).__init__()
        if config_files:
            self.config_files = tuple(config_files)
        else:
            self.config_files = self.default.config_files

    def setup_fixture(self):
        self.setup_ssh_config()

    def setup_ssh_config(self):
        self.config = paramiko.SSHConfig()
        for config_file in self.config_files:
            config_file = os.path.expanduser(config_file)
            if os.path.exists(config_file):
                LOG.debug("Parsing %r config file...", config_file)
                with open(config_file) as f:
                    self.config.parse(f)
                LOG.debug("File %r parsed.", config_file)

    def lookup(self, host=None):
        host_config = host and self.config.lookup(host) or {}
        # remove unsupported directive
        host_config.pop('include', None)
        return SSHHostConfig(host=host,
                             ssh_config=self,
                             host_config=host_config,
                             config_files=self.config_files)

    def __repr__(self):
        return "{class_name!s}(config_files={config_files!r})".format(
            class_name=type(self).__name__, config_files=self.config_files)


class SSHHostConfig(collections.namedtuple('SSHHostConfig', ['host',
                                                             'ssh_config',
                                                             'host_config',
                                                             'config_files'])):

    default = tobiko.required_setup_fixture(SSHDefaultConfigFixture)

    @property
    def hostname(self):
        return self.host_config.get('hostname', self.host)

    @property
    def port(self):
        return (self.host_config.get('port') or
                self.default.port)

    @property
    def username(self):
        return (self.host_config.get('user') or
                self.default.username)

    @property
    def key_filename(self):
        return (self.host_config.get('identityfile') or
                self.default.key_file)

    @property
    def proxy_jump(self):
        proxy_jump = (self.host_config.get('proxyjump') or
                      self.default.proxy_jump)
        if not proxy_jump:
            return None

        proxy_hostname = self.ssh_config.lookup(proxy_jump).hostname
        if ({proxy_jump, proxy_hostname} & {self.host, self.hostname}):
            # Avoid recursive proxy jump definition
            return None

        return proxy_jump

    @property
    def proxy_command(self):
        return (self.host_config.get('proxycommand') or
                self.default.proxy_command)

    @property
    def allow_agent(self):
        return (is_yes(self.host_config.get('forwardagent')) or
                self.default.allow_agent)

    @property
    def compress(self):
        return (is_yes(self.host_config.get('compression')) or
                self.default.compress)

    @property
    def timeout(self):
        return (self.host_config.get('connetcttimeout') or
                self.default.timeout)

    @property
    def connection_attempts(self):
        return (self.host_config.get('connectionattempts') or
                self.default.connection_attempts)

    @property
    def connection_interval(self):
        return (self.host_config.get('connetcttimeout') or
                self.default.connection_interval)

    @property
    def connect_parameters(self):
        return dict(hostname=self.hostname,
                    port=self.port,
                    username=self.username,
                    key_filename=self.key_filename,
                    compress=self.compress,
                    timeout=self.timeout,
                    allow_agent=self.allow_agent,
                    connection_attempts=self.connection_attempts,
                    connection_interval=self.connection_interval)


def is_yes(value):
    return value and str(value).lower() == 'yes'



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko


class ShellError(tobiko.TobikoException):
    pass


class ShellCommandFailed(ShellError):
    """Raised when shell command exited with non-zero status
    """
    message = ("command {command!r} failed (exit status is {exit_status}); "
               "stderr:\n{stderr!s}\n"
               "stdout:\n{stdout!s}")


class ShellTimeoutExpired(ShellError):
    """Raised when shell command timeouts and has been killed before exiting
    """
    message = ("command {command!r} timed out after {timeout!s} seconds; "
               "stderr:\n{stderr!s}\n"
               "stdout:\n{stdout!s}")
